04/15/97 12:39:51 PM

for: aScanner
	" Returns an instance of the receiver working on aScanner. "

	^self new
		scanner: aScanner;
		yourself! 
comment
    " Documentation - Returns the receiver's comment string. "

    ^'Implements support for Twain Scanner Objects'!   
fileTransferFormat
    "Returns the instance var of the receiver
     for external manipulation"

	fileTransferFormat isNil ifTrue: [ self initializeFileTransferFormat ].
    ^fileTransferFormat!  
initializeFileTransferPrologue
    "Private - Initialize transfer prologue with default values."

	fileTransferPrologue := super fileTransferPrologue! 
useBitmapFileFormat
	"Use Windows Bitmap during file transfer of images."

	self fileTransferFormat: TwffBmp!  
initializeFileTransferFormat
    "Private - Initialize transfer format with default values."

	fileTransferFormat := super fileTransferFormat! 
fileTransferPrologue
    "Returns the instance var of the receiver
     for external manipulation"

	fileTransferPrologue isNil ifTrue: [ self initializeFileTransferPrologue ].
    ^fileTransferPrologue!  
useJfifFileFormat
	"Use JPEG File Interchange Format during file transfer of images."

	self fileTransferFormat: TwffJfif! 
fileTransferExtension
	" Private - Returns a file name extension of a new image."

	| format |
	format := self fileTransferFormat.
	format == TwffBmp ifTrue: [ ^'.BMP' ].
	format == TwffTiff ifTrue: [ ^'.TIF' ].
	format == TwffJfif ifTrue: [ ^'.JPG' ].
	format == TwffPict ifTrue: [ ^'.PIC' ].
	format == TwffXbm ifTrue: [ ^'.XBM' ].
	^super fileTransferExtension!
fileTransferPrologue: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    fileTransferPrologue := anObject! 
usePictFileFormat
	"Use Macintosh PICT format during file transfer of images."

	self fileTransferFormat: TwffPict!
useTiffFileFormat
	"Use Tagged Image File Format during file transfer of images."

	self fileTransferFormat: TwffTiff! 
fileTransferFormat: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    fileTransferFormat := anObject! 
useFileTransfer
	"Use file transfer mechanism."

	self transferMechanism: TwsxFile!
useNativeTransfer
	"Use native transfer mechanism."

	self transferMechanism: TwsxNative!  
useXbmFileFormat
	"Use X-Windows Bitmap format during file transfer of images."

	self fileTransferFormat: TwffXbm!
initializeTransferMechanism
    "Private - Initialize transfer mechanism with default values."

	transferMechanism := super transferMechanism! 
useDefaultTransfer
	"Use default transfer mechanism."

	self transferMechanism: nil!   
transferMechanism
    "Returns the instance var of the receiver
     for external manipulation"

	transferMechanism isNil ifTrue: [ self initializeTransferMechanism ].
    ^transferMechanism!  
transferMechanism: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    transferMechanism := anObject!   
useMemoryTransfer
	"Use memory transfer mechanism."

	self transferMechanism: TwsxMemory!  
fileTransferFormat
	"Returns the default file transfer format."

	^TwffBmp!
transferMechanism
	"Returns the transfer mechanism."

	^self scanner transferMechanism!
fileTransferExtension
	" Private - Returns a file name extension of a new image."

	^'.bmp'!   
scanner
    "Returns the instance var of the receiver
     for external manipulation"
    ^scanner!
scanner: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    scanner := anObject!   
fileTransferFileName
	"Returns a file name for next file transfer of an image."

	^((self fileTransferPrologue ,Time now asSeconds asString,'        ')
		copyFrom: 1 to: 8) trimBlanks, self fileTransferExtension!  
fileTransferPrologue
	" Private - Returns a prologue string for file name of a new image."

	^'tw'!
countryNamed: aName
	" Private - Returns the id for country named aName. "

	^self twainDLLClass countryNamed: aName!  
structureNamed: aName atAddress: anAddress
	" Private - Returns the registered structure named aName. "

	^self twainDLLClass structureNamed: aName atAddress: anAddress!  
newAppIdentityFamily: familyString name: nameString
    "Returns an AppIdentity structure or nil if failed...
    The structure will be initialized acording to
     local settings..."

    | struct stVersion |
    struct := self structureNamed: 'TW_IDENTITY'.
    struct notNil ifTrue: [
        stVersion := ReadStream on: Smalltalk version.
        struct version
            majorNum: (stVersion upTo: $.) asInteger;
            minorNum: (stVersion upTo: nil) asInteger;
            language: TwlgSpa;
            country: (self countryNamed: 'Argentina');
            info: '@Object Style Application';
            yourself.
        struct
            id: 0; "init to 0, Source Manager will asign real value"
            protocolMajor: TwonProtocolmajor;
            protocolMinor: TwonProtocolminor;
            supportedGroups: DgImage | DgControl;
            manufacturer: 'Object Style';
            productFamily: familyString;
            productName: nameString;
            yourself.
    ].
    ^struct!   
newAppIdentity
    "Returns an AppIdentity structure or nil if failed...
    The structure will be initialized acording to
     local settings..."

    | default |
    default := 'Object Style'.
    ^self newAppIdentityFamily: default name: default!  
twainDLLClass
	" Private - Returns the low level support for twain dynalink. "

	^Twain32DLL!  
comment
    " Documentation - Returns the receiver's comment string. "

    ^'Implements Twain Scanner support.'!  
test
    "Test of the interfase.
    TwainScanner test
    "

    | scanner result acquireBlock |
    acquireBlock := [:aTwainResult|
        ((Notifier respondsTo: #isControlKeyDown) and: [ Notifier isControlKeyDown ])
        ifTrue: [ aTwainResult halt ].
        aTwainResult asDIB inspect.
    ].
    scanner := self new.
    result  := scanner selectSource.
    result isNil ifTrue: [ ^self "cancelled" ].
    result ifTrue: [ scanner acquireDoing: acquireBlock ]
    ifFalse: [
        MessageBox warning: 'Interfz TWAIN inoperante o seleccin de dispositivo cancelada'
    ].
    scanner release.!  
structureNamed: aName
	" Private - Returns the registered structure named aName. "

	^self twainDLLClass structureNamed: aName!
dsmEntry: pOrigin pDest: pDest dg: dg dat: dat msg: msg pData: pData
    "Private - DSM_Entry, the only entry point into the Data Source Manager...
	DEBUG NOTE: we implement this message here for easy debuggin.
    Parameters:
        pOrigin Identifies the source module of the message.
                    This could identify an application, a Source, or the Source Manager.
        pDest   Identifies the destination module for the message.
                    This could identify an application or a data source.
                    If this is NULL, the message goes to the Source Manager.
        DG      The Data Group. Example: DG_IMAGE.
        DAT     The Data Attribute Type. Example: DAT_IMAGEMEMXFER.
        MSG    The message.  Messages are interpreted by the destination module
                    with respect to the Data Group and the Data Attribute Type.
                    Example: MSG_GET.
        pData   A pointer to the data structure or variable identified
                    by the Data Attribute Type.
                    Example: (TW_MEMREF)&ImageMemXfer
                         where ImageMemXfer is a TW_IMAGEMEMXFER structure.
    Returns: ReturnCode  Example: DCRC_SUCCESS.
    "
	^self twainDLL dsmEntry: pOrigin pDest: pDest dg: dg dat: dat msg: msg pData: pData!   
isDSOpen
    "Private - True if the Data Source is currently open"

    ^dsOpen!   
errorBlock: oneArgumentBlock
    "Sets a block to be used to handle errors in the receiver.
    aBlock must be a one-param block to be valued with an errorString.
    aBlock can be nil for default behavior."

    errorBlock := oneArgumentBlock! 
initializeTwainDLL
	" Private - Gets the low level twain DLL. "

	twainDLL := self class twainDLLClass current.!   
getCurrentCapability: aCapConstant
    "Returns the current capability value(/values) for aCapConstant.
    aCapConstants must be anInteger ICAPxx or Capxx constant value..."

    ^self getCapability: aCapConstant withMessage: MsgGetcurrent! 
getDefaultCapability: aCapConstant
    "Returns the default capability value(/values) for aCapConstant.
    aCapConstants must be anInteger ICAPxx or Capxx constant value..."

    ^self getCapability: aCapConstant withMessage: MsgGetdefault! 
getCompleteImageMemory
    "Private - Ask the Source to start sending the pending image in direct memory transfer mode..."

    | dcRC xfer setup theMem buffer bufferSize freeMem result |
	result := TwainMemoryResult for: self.
    xfer := self class structureNamed: 'TW_IMAGEMEMXFER'.
    setup := self memoryTransferProperties.
    bufferSize := setup preferred.
    theMem := self allocateMemory: bufferSize.
    freeMem := [ self releaseMemory: theMem. ].
    [
        xfer memory
            flags: TwmfAppowns | TwmfPointer;
            length: bufferSize;
            theMem: theMem asParameter;
            yourself.
        dcRC := self
            dsmEntry: self appIdentity asParameter
            pDest: self currentDataSource asParameter
            dg: DgImage
            dat: DatImagememxfer
            msg: MsgGet
            pData: xfer asParameter.
        (dcRC = TwrcSuccess or: [dcRC = TwrcXferdone])
        ifTrue: [
			result addBytes: xfer bytesWritten from: theMem.
        ]
        ifFalse: [
            dcRC = TwrcCancel
            ifTrue: [ "the user canceled or wants to rescan the image"
				result cancelled.
			]
            ifFalse: [ "Something wrong, abort the transfer"
                self explainError: 'Something wrong, abort the transfer' number: dcRC.
                (MessageBox confirm: 'Retry operation ?') ifFalse: [
					result aborted.
					freeMem value.
					^nil
				].
            ]
        ].
    dcRC = TwrcXferdone ] whileFalse: [].
    freeMem value.
	result completed.
	self acquired: result.!  
disableDS
    "Private - Hide DataSource User Interfase..."

    | dcRC dcUI |
    self isDSOpen ifFalse: [ ^self ].
    self openDS ifFalse: [ ^self ].
    dcUI := self class structureNamed: 'TW_USERINTERFACE'.
    dcUI
        hParent: self currentUIParent handle;
        showUI: 0.
    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: self currentDataSource asParameter
        dg: DgControl
        dat: DatUserinterface
        msg: MsgDisableds
        pData: dcUI asParameter.

    dcRC = TwccSuccess ifTrue: [^self].
    self explainError: 'Error when closing Data Source User interface' number: dcRC!   
isExternalAddress: anObject
    " Private - Return true if anObject is an external address. "

    ^anObject isKindOf: self externalAddressClass!  
processDSEvent: winEvent
    "Private - For TwainWindow message loop processing..."

    | dcEvent dcRC dsMsg |

    (self isDSMOpen and: [self isDSOpen]) ifFalse: [^false].

    "A Source provides a modeless dialog box as its user interface.
     The following call relays Windows messages down to the Source's
     UI that were intended for its dialog box.  It also retrieves TWAIN
     messages sent from the Source to our Application."

    dcEvent := self class structureNamed: 'TW_EVENT'.
    dcEvent pEvent: winEvent asParameter.

    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: self currentDataSource asParameter
        dg: DgControl
        dat: DatEvent
        msg: MsgProcessevent
        pData: dcEvent asParameter.

    dcRC = TwrcDsevent ifFalse: [^false].

    dsMsg := dcEvent dcMessage.
    dsMsg = MsgXferready ifTrue: [ self transferImage ]
    ifFalse: [
        dsMsg = MsgClosedsreq ifTrue: [ self dcTerminate ]
        ifFalse: [
            dsMsg = MsgNull ifTrue: [ "No message from the Source to the App" ]
            ifFalse: [ "unKnown message received..."
                ('unKnown Twain message received #',dsMsg asString) displayAt: 10@10
            ].
        ].
    ].

    ^true! 
externalAddressClass
    " Private - Return the class for external address support. "

    #platformDependent.
    ^Smalltalk at: #WinAddress
        ifAbsent: [ Smalltalk at: #ExternalAddress ]!  
supportedCapabilities
    "Returns the supported Capabilities flag values.
	WARNING: Capability can be unsupported by some sources.
	Returns values or nil if unsupported."

    ^self getCapability: CapSupportedcaps!  
enableDS
    "Private - Brings up DataSource User Interfase...
    Returns : true if open false if failed..."

    | dcRC dcUI |
    self openDS ifFalse: [ ^false ].
    dcUI := self class structureNamed: 'TW_USERINTERFACE'.
    dcUI
        hParent: self currentUIParent handle;
        showUI:  16rFFFF.
    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: self currentDataSource asParameter
        dg: DgControl
        dat: DatUserinterface
        msg: MsgEnableds
        pData: dcUI asParameter.

    dcRC = TwccSuccess ifTrue: [^true].
    self explainError: 'Error when opening Data Source User interface' number: dcRC.
    ^false!   
currentDataSource
    "Private - Returns the current DS in use,
    the default if not used yet,
    or nil if there are any problem...
    Note that the DS is an AppIdentity structure object."

    dataSourceID isNil ifTrue: [
        dataSourceID := self defaultDataSource
    ].

    ^dataSourceID!   
defaultDataSource
    "Private - Query the default data source...
    WARNING : DSM must be open..."

    | dcRC newIdentity |
    newIdentity := self class newAppIdentity.
    "fill newIdentity with default source settings..."
    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: nil
        dg: DgControl
        dat: DatIdentity
        msg: MsgGetdefault
        pData: newIdentity asParameter.

    dcRC = TwccSuccess
    ifFalse: [
        self explainError: 'Error when User select Source' number: dcRC.
        ^nil "failed..."
    ].

    ^newIdentity!  
openDS
    "Private - Open the only Data Source...
    Returns : current state of the select Source"

    | dcRC |
    self openDSM ifFalse: [ ^dsOpen := false ].
    dsOpen == true ifTrue: [^true].

    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: nil
        dg: DgControl
        dat: DatIdentity
        msg: MsgOpends
        pData: self currentDataSource asParameter.

   dsOpen := dcRC = TwccSuccess.

    dsOpen ifFalse: [
        self explainError: 'Error when opening Data Source' number: dcRC
    ].

    ^dsOpen!  
selectDS
    "Private - Call the Source Manager to:
        - have SM put up a list of available Sources.
        - get information about the user selected Source
            from NewDSIdentity, filled by Source...
    Returns true if success else return false
        or nil if cancelled by the user...
    "

    | dcRC newIdentity |
    newIdentity := self defaultDataSource.
    "open the SM selectSource dialog box..."
    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: nil
        dg: DgControl
        dat: DatIdentity
        msg: MsgUserselect
        pData: newIdentity asParameter.
    dcRC = TwrcCancel ifTrue: [ ^nil ].
    dcRC = TwrcSuccess
    ifTrue: [ dataSourceID := newIdentity ]
    ifFalse: [ self explainError: 'Error when User select Source' number: dcRC ].
    ^ dcRC = TwrcSuccess!
status
    "Private - Returns an status structure or nil if error..."

    | status dcRc |
    status := self class structureNamed: 'TW_STATUS'.
    dcRc := self
        dsmEntry: self appIdentity asParameter
        pDest: nil
        dg: DgControl
        dat: DatStatus
        msg: MsgGet
        pData: status asParameter.
    dcRc = TwccSuccess ifFalse: [^nil]. "no DSM..."
    ^status! 
extractEnumerationValue: dcCap
    "Private - Returns an enumetation value from a dcCap structure..."

    | enum result itemList items areBooleans |
    enum := self class
		structureNamed: 'TW_ENUMERATION'
		atAddress: (self globalLock: dcCap hContainer).
    items := enum itemList.
    self setTypesOf: items with: enum itemType.
	areBooleans := enum itemType == TwtyBool.

    result := OrderedCollection new.
    1 to: enum numItems do: [:i|
        result add: (items at: i)
    ].
	areBooleans ifTrue: [
		result := result collect: [:each| each ~= 0 ]
	].

    self globalUnlock: dcCap hContainer.
    ^result asArray!
units
    "Returns the supported units.
	WARNING: Capability can be unsupported by some sources.
	Returns a collection of values or nil if unsupported."

    ^self getCapability: IcapUnits!
explainError: aString number: dcRC
    "Private - Process the error notification with an error constant number.."

    ^self explainError: aString
        ,'. dcRC=',(self constantWithPrefix: 'Twrc'
                            in: TwainConstants
                            withValue: dcRC
                            ifAbsent: [ dcRC asString ])!
peekMessage: aBoolean
    "Private - Call UserLibrary peekMesage...
    Doing as is for compatibility reasons"

    "WARNING: VM dependent method..."
	#platformDependent.

    ^Smalltalk version = '2.0'
    ifTrue: [
        UserLibrary
            peekMessage: WinMsgNS asParameter
            hwnd: nil min: 0 max: 0
            remove: aBoolean
    ]
    ifFalse: [
        UserLibrary
            peekMessage: WinMsgNS asParameter
            hwnd: nil min: 0 max: 0
            remove: aBoolean asParameter
    ]! 
memoryTransferProperties
    "Private - Returns the current memory transfer settings."

    | setup dcRC |
    setup := self class structureNamed: 'TW_SETUPMEMXFER'.
    setup "fill with stupid values because some scanners failed when this struct is filled with zeros"
        maxBufSize: 16rFFFF;
        minBufSize: 0;
        preferred: 16 * 1024;
        yourself.

    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: self currentDataSource asParameter
        dg: DgControl
        dat: DatSetupmemxfer
        msg: MsgGet
        pData: setup asParameter.

    dcRC = TwrcSuccess ifTrue: [ ^setup ].
    "self explainError: 'Error in #memoryTransferProperties' number: dcRC."
    ^nil!   
release
    "Private - Release the receiver. "

    "Close SM & DSM this is required by ''the rules''..."
    self
        dcTerminate;
        initialize.

    ^super release.! 
transferImage
    "Private - Transferr the captured image..."

    | dcImageInfo dcPendingXfer dcRC |

    dcImageInfo := self class structureNamed: 'TW_IMAGEINFO'.

    "Tell the Source what type of transfer you want.  I use the default,
      NATIVE so no set of ICAP_XFERMECH is required "

    " Native also negates the need to ask for a pixel type ICAP_PIXELTYPE...
      self pixelType: TwptRgb.
      pixelType := self pixelType.
    "

    "Get the image information, nice to know a little about the image
      the Source will be sending..."
    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: self currentDataSource asParameter
        dg: DgImage
        dat: DatImageinfo
        msg: MsgGet
        pData: dcImageInfo asParameter.

    dcRC = TwccSuccess ifTrue: [
        dcPendingXfer := self class structureNamed: 'TW_PENDINGXFERS'.
        dcPendingXfer count: 0.
        [
		  CursorManager execute changeFor: [
            self getCompleteImage isNil
            ifTrue: [ dcPendingXfer count: 0 ] "failed..."
            ifFalse: [
                dcRC := self "Required for proper 6<->7 state transitions"
                    dsmEntry: self appIdentity asParameter
                    pDest: self currentDataSource asParameter
                    dg: DgControl
                    dat: DatPendingxfers
                    msg: MsgEndxfer
                    pData: dcPendingXfer asParameter.
                dcRC = TwccSuccess ifFalse: [ dcPendingXfer count: 0 ]. "failed"
            ].
		  ].
        dcPendingXfer count = 0 ] whileFalse: [].
    ].

    "shut down the Source..."
    self dcTerminate.!  
imageFileFormats
    "Returns the posible image File Formats.
	WARNING: Capability can be unsupported by some sources.
	Returns a collection of values or nil if unsupported."

    ^self getCapability: IcapImagefileformat!
getCompleteImageNative
    "Private - Ask the Source to start sending the pending image in native transfer mode."

    | hBitmap dcRC result |
    hBitmap := self externalLongClass new.
	result := TwainNativeResult for: self.
    [   dcRC := self
            dsmEntry: self appIdentity asParameter
            pDest: self currentDataSource asParameter
            dg: DgImage
            dat: DatImagenativexfer
            msg: MsgGet
            pData: hBitmap asParameter.
        dcRC = TwrcXferdone
        ifTrue: [
			result hBitmap: hBitmap.
			self acquired: result.
			"self globalFree: hBitmap asParameter."
		]
        ifFalse: [
            dcRC = TwrcCancel
            ifTrue: [ "the user canceled or wants to rescan the image" ]
            ifFalse: [ "Something wrong, abort the transfer"
                self explainError: 'Something wrong, abort the transfer' number: dcRC.
                (MessageBox confirm: 'Retry operation ?') ifFalse: [ ^nil ]. "abort"
            ]
        ].
    dcRC = TwrcXferdone ] whileFalse: [].
	^result!
globalUnlock: aHandle
    "Private - Global unlock of a handle (PUAJ)..."

    KernelLibrary globalUnlock: aHandle!
errorBlock
    "Returns the instance var of the receiver
     for external manipulation"
    ^errorBlock!  
closeDSM
    "Private - Close the Data Source Manager.
    Returns : Current state of the Source Manager
    Notes:
        1) be sure SM is already open
        2) explicity load the .DLL for the Source Manager
        3) call Source Manager to:
                - request closure of the Source identified by appID info
    "
    | dcRC |
    self isDSMOpen ifFalse: [ ^dsmOpen ].

    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: nil
        dg: DgControl
        dat: DatParent
        msg: MsgClosedsm
        pData: self currentUIParent asParameter.

    "No calienta...
    dcRC = TwccSuccess ifFalse: [
        self explainError: 'Error when closing Data Source Manager'
    ].
    "

    ^dsmOpen := false! 
closeDS
    "Private - Close the current Data Source..."

    | dcRC |
    self isDSOpen ifFalse: [ ^self ].

    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: nil
        dg: DgControl
        dat: DatParent
        msg: MsgCloseds
        pData: self currentDataSource asParameter.

    dsOpen := false.

    "No calienta...
    dcRC = TwccSuccess ifFalse: [
        self explainError: 'Error when closing Data Source'
    ].
    "!   
acquire: parentWindow doing: aBlock
    "Acquire now...& evaluate aBlock with each acquired chunk of information.
    parentWindow will be used as the parent window for the UI dialog box"

    | oldAcq |
    oldAcq := acquireBlock.
    acquireBlock := aBlock.
    self acquire: parentWindow.
    acquireBlock := oldAcq.!  
readWinQueue
    "Private - Read a message from host system's event queue.
        Puajjjj da asco pero no queda otra..."

    | old |
    #platformDependent.
    old :=  Process enableInterrupts: false.
    (self peekMessage: false)
    ifTrue: [
        (self processDSEvent: WinMsgNS)
        ifFalse: [ Notifier readWinQueue. ]
        ifTrue: [
            (self peekMessage: true)
            ifTrue: [
                UserLibrary
                    translateMessage: WinMsgNS asParameter;
                    dispatchMessage: WinMsgNS asParameter.
            ].
        ].
    ].
    Process enableInterrupts: old!   
dcTerminate
    "Private - Close all interfases & window."

      self
        disableDS;
        closeDS;
        closeDSM.!   
isValidWindow: aWindow
    " Private - Return true if aWindow is valid and open. "

    ^aWindow isWindow and: [ aWindow isHandleOk]!  
pixelType
    "Returns the current pixel type.
	WARNING: Capability can be unsupported by some sources.
	Returns a collection of values or nil if unsupported."

    ^self getCurrentCapability: IcapPixeltype!  
pixelTypes
    "Returns the suported pixel types.
	WARNING: Capability can be unsupported by some sources.
	Returns a collection of values or nil if unsupported."

    ^self getCapability: IcapPixeltype!  
selectSource
    "Lets the user select the data source.
    Returns :
        true - if success
        false - if DSM or DS failed
        nil - if the user select cancel...
    "

    | response focused |
	focused := self activeMainWindow.

    self openDSM
    ifTrue: [
        response := self selectDS.
        self closeDSM.
		focused notNil ifTrue: [ focused setFocus ].
        ^response
    ].
	focused notNil ifTrue: [ focused setFocus ].
    ^false!  
transferMechanism
    "Returns the current image transfer mechanism.
	WARNING: Capability can be unsupported by some sources.
	Returns a collection of values or nil if unsupported."

    ^self getCurrentCapability: IcapXfermech! 
transferMechanisms
    "Returns the suported image transfer mechanisms.
	WARNING: Capability can be unsupported by some sources.
	Returns a collection of values or nil if unsupported."

    ^self getCapability: IcapXfermech! 
getCapability: aCapConstant
    "Returns the capability value(/values) for aCapConstant.
    aCapConstants must be anInteger ICAPxx or Capxx constant value..."

    ^self getCapability: aCapConstant withMessage: MsgGet!   
extractArrayValue: dcCap
    "Private - Returns an enumetation value from a dcCap structure..."

    | array result itemList items areBooleans |
    array := self class
		structureNamed: 'TW_ARRAY'
		atAddress: (self globalLock: dcCap hContainer).
    items := array itemList.
    self setTypesOf: items with: array itemType.
	areBooleans := array itemType == TwtyBool.

    result := OrderedCollection new.
    1 to: array numItems do: [:i|
        result add: (items at: i)
    ].
	areBooleans ifTrue: [
		result := result collect: [:each| each ~= 0 ]
	].

    self globalUnlock: dcCap hContainer.
    ^result asArray!  
feederLoaded
    "Returns the feederLoaded property.
	WARNING: Capability can be unsupported by some sources.
	Returns a boolean or nil if unsupported."

    ^self getCapability: CapFeederloaded!  
isDSMOpen
    "Private - True if the Data Source Manager is currently open"

    dsmOpen isNil ifTrue: [ self initialize ].
    ^dsmOpen! 
acquire
    "Acquire now..."

    ^self acquire: nil!  
acquire: parentWindow
    "Acquire now..."
    "Note that the DS User interfase will be opened until
     disabled by a message received in receiver's window trapper."

    | debuggin |
    debuggin := false. "*DEBUG* Only used for debuggin"
	parentUI := parentWindow.

    self openDS ifFalse: [^false]
    ifTrue: [
        self enableDS
        ifFalse: [self closeDS.^false].
    ].

    self consumeInputUntil: [:event|
        self isDSOpen not"UI open"
        or: [ debuggin and: [Cursor sense x < 100]]
    ].

    self isDSOpen ifTrue: [
        debuggin ifTrue: [ Terminal bell. Terminal cr;nextPutAll: 'Debug UI Abort...' ].
        self dcTerminate
    ]. "when aborted during debuggin..."

	parentWindow notNil ifTrue: [ parentWindow setFocus ].
    ^true!  
appIdentity
    "Private - Returns the current AppIdentity in use (or nil)"

   appIdentity isNil ifTrue: [
        appIdentity := self class newAppIdentity
    ].

    ^appIdentity! 
dsEntry: pOrigin dg: dg dat: dat msg: msg pData: pData
    "Private - DS_Entry, the entry point provided by a Data Source...
	DEBUG NOTE: we implement this message here for easy debuggin.
    Parameters:
        pOrigin Identifies the source module of the message.
                    This could identify an application or the Data Source Manager.
        DG      The Data Group. Example: DG_IMAGE.
        DAT     The Data Attribute Type. Example: DAT_IMAGEMEMXFER.
        MSG    The message.  Messages are interpreted by the data source
                    with respect to the Data Group and the Data Attribute Type.
                    Example: MSG_GET.
        pData   A pointer to the data structure or variable identified
                    by the Data Attribute Type.
                    Example: (TW_MEMREF)&ImageMemXfer
                         where ImageMemXfer is a TW_IMAGEMEMXFER structure.
    Returns: ReturnCode  Example: DCRC_SUCCESS.

    Note:
        The DSPROC type is only used by an application when it calls
        a Data Source directly, bypassing the Data Source Manager.

    "
	^self twainDLL dsEntry: pOrigin dg: dg dat: dat msg: msg pData: pData!
getCompleteImage
    "Private - Ask the Source to start sending the pending image..."

	| transferMechanism |
	transferMechanism := self settings transferMechanism.

	transferMechanism= TwsxNative
	ifTrue: [ ^self getCompleteImageNative ].

	transferMechanism = TwsxFile
	ifTrue: [ ^self getCompleteImageFile ].

	transferMechanism = TwsxMemory
	ifTrue: [ ^self getCompleteImageMemory ].

	^self getCompleteImageNative "use native if unknown transfer mechanism"! 
acquireDoing: aBlock
    "Acquire now!!... & evaluate aBlock with each acquired twainResult"

    ^self acquire: self activeMainWindow doing: aBlock!  
allocateMemory: sizeInBytes
	" Private - Returns a new external block of memory of size sizeInBytes. "

	^self externalAddressClass allocateMemory: sizeInBytes!   
releaseMemory: anExternalMemory
	" Private - Frees a new external block of memory. "

	anExternalMemory isNil ifTrue: [ ^self ].

	#platformDependent.
	(anExternalMemory respondsTo: #unlockAndFree)
	ifTrue: [ anExternalMemory unlockAndFree ]
	ifFalse: [ anExternalMemory free ].!   
getCapability: aCapConstant withMessage: aMsgGetConstant
    "Private - Returns the capability value for aCapConstant.
    aCapConstants must be anInteger ICAPxx or Capxx constant value..."

    | dcRC dcCap result |
    dcCap := self class structureNamed: 'TW_CAPABILITY'.
    dcCap cap: aCapConstant.

    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: self currentDataSource asParameter
        dg: DgControl
        dat: DatCapability
        msg: aMsgGetConstant
        pData: dcCap asParameter.

    dcRC = TwccSuccess ifFalse: [
        self explainError: 'During Capability request...' number: dcRC.
        ^nil
    ].

    dcCap conType = TwonOnevalue ifTrue: [ result := self extractOneValue: dcCap ].
    dcCap conType = TwonEnumeration ifTrue: [ result := self extractEnumerationValue: dcCap ].
    dcCap conType = TwonArray ifTrue: [ result := self extractArrayValue: dcCap ].

    self globalFree: dcCap hContainer.
    ^result!
acquired: aTwainResult
	" Private - Process aTwainResult performing acquire block on it. "

	aTwainResult isNil ifTrue: [ ^self ].

    (acquireBlock respondsTo: #value:) ifTrue: [
        ^acquireBlock value: aTwainResult
    ].

    acquireBlock isMessage ifTrue: [
        ^acquireBlock
            arguments: (Array with: aTwainResult);
            perform
    ].

    ^self explainError: 'Twain do not know what to do with catured images.'!  
pixelFlavor
    "Returns the pixelFlavor
	WARNING: Capability can be unsupported by some sources.
	Returns a value or nil if unsupported.
	zero pixel represents darkest shade for TwpfChocolate
		and lightest for TwpfVanilla.
	"

    ^self getCurrentCapability: IcapPixelflavor! 
initialize
    "Private - Initializes the receiver..."

	super initialize.
    dsmOpen := dsOpen := false.
    dataSourceID := appIdentity := nil.!  
settings
    "Returns the instance var of the receiver for external manipulation."

	settings isNil ifTrue: [ self initializeSettings ].
    ^settings!   
globalFree: aHandle
    "Private - Global free of a handle (PUAJ)..."

    " MessageBox message: 'Global Free :',aHandle asString. "
    KernelLibrary globalFree: aHandle!   
globalLock: aHandle
    " Private - Returns the an external address object from aHandle locking aHandle.
	WARNING: the returned addresss MUST be freed. "

	| addressPtr |
    #platformDependent.
    (Smalltalk includesKey: #ExternalAddress)
	ifTrue: [	"32 bits version"
	    (addressPtr :=  KernelLibrary globalLock: aHandle asParameter) = 0 ifTrue: [ ^self osError ].
    	^(Smalltalk at: #ExternalAddress) fromInteger: addressPtr
	]
	ifFalse: [	"16 bits version"
	    addressPtr := KernelLibrary globalLock: aHandle asParameter.
    	^(Smalltalk at: #WinAddress) fromString: addressPtr
    ].! 
settings: aTwainSetting
    "Sets the instance var of the receiver."

	aTwainSetting scanner: self.
    settings := aTwainSetting!
constantWithPrefix: prefix in: aPoolDictionary withValue: aValue ifAbsent: aBlock
    "Private - Returns the name of a pool dictionary constant
    of value aValue and prefix prefix (if prefix isEmpty all the pool
    dictionary will be used for searching).
    aBlock will be evaluated if not found"

    aPoolDictionary associationsDo: [:asoc|
        (prefix size = 0 or: [(asoc key indexOfCollection: prefix) > 0])
        ifTrue: [
            aValue = asoc value
            ifTrue: [ ^asoc key ]
        ]
    ].
    ^aBlock value!   
explainError: aString
    "Private - Process the error notification.."

    | errors errorStr status |
    errors := Dictionary new.
    errors
        at: TwccNods put: 'No TWAIN Source found in the TWAIN directory.\  Refer to your TWAIN Source installation instructions.';
        at: TwccLowmemory put: 'There is not enough memory to Acquire.';
        at: TwccMaxconnections put: 'The TWAIN Source is already in use.\Data Source is connected to max possible applications.';
        at: TwccBummer put: 'Failure due to unknown causes.';
        at: TwccBadcap put: 'Capability error has occurred.';
        at: TwccBadprotocol put: 'Protocol error has occurred.';
        at: TwccBadvalue put: 'Bad value.';
        at: TwccOperationerror put: 'Operation Error';
        at: TwccSeqerror put: 'DG DAT MSG is out of expected sequence.';
        at: TwccBaddest put: 'Unknown destination Application/Source in DSM_Entry';
        yourself.
    status := self status.
    errorStr := status isNil
        ifTrue: [ 'Could not get detailed Error Code.' ]
        ifFalse: [
            status code = 0 ifTrue: [ 'Status signal indicates no error.' ]
            ifFalse: [
                errors
                    at: status code
                    ifAbsent: [ '* UnKnow error : ',status code asString,' *' ]
            ]
        ].
    errorStr := (errorStr,'\',aString) withCrs trimBlanks.
    errorStr size = 0 ifTrue: [^self]. "don't care..."

    errorBlock isNil ifTrue: [
        MessageBox notify: 'Twain Error' withText: errorStr.
    ]
    ifFalse: [ errorBlock value: errorStr ].!  
setTypesOf: items with: itemType
    "Private - Set the proper items tyepes to an open SDS array..."

    itemType = TwtyInt8 ifTrue: [ items itemType: #byte ].
    itemType = TwtyUint8 ifTrue: [ items itemType: #byte ].
    itemType = TwtyInt16 ifTrue: [ items itemType: #short ].
    itemType = TwtyUint16 ifTrue: [ items itemType: #ushort ].
    itemType = TwtyInt32 ifTrue: [ items itemType: #long ].
    itemType = TwtyUint32 ifTrue: [ items itemType: #ulong ].
    itemType = TwtyBool ifTrue: [ items itemType: #ushort ].! 
initializeSettings
    "Private - Initialize receiver's settings."

	settings := TwainDefaultSettings for: self.!  
externalLongClass
    " Private - Return the class for external long support. "

    #platformDependent.
    ^Smalltalk at: #WinLong
        ifAbsent: [ Smalltalk at: #ExternalLong ]!  
fileTransferProperties: setup
    "Private - Sets the current file transfer parameters."

    | dcRC |
    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: self currentDataSource asParameter
        dg: DgControl
        dat: DatSetupfilexfer
        msg: MsgSet
        pData: setup asParameter.

    dcRC = TwrcSuccess ifTrue: [ ^true ].
    self explainError: 'Error in setting fileTransferProperties' number: dcRC.
    ^false!
activeMainWindow
	" Private - Returns current active window or nil."

	#platformDependent.
	^Notifier activeMainWindow!   
twainDLL
	" Private - Returns the low level twain DLL in use. "

	twainDLL isNil ifTrue: [
		self initializeTwainDLL
	].
	^twainDLL!
consumeInputUntil: aBlock
    "Private - Consume Win Events for catching DS-Events.
        Puajjjj da asco pero no queda otra..."

    "WARNING: VM dependent method...
    Copied from NotificationManager...
    see changes in self implementation of #readWinQueue..."

    | e int |
    #platformDependent.
    [CurrentEvents isEmpty] whileTrue: [ self readWinQueue].
    int := Process enableInterrupts: false.
    e := CurrentEvents removeFirst.
    Process enableInterrupts: int.
    [aBlock value: e] whileFalse: [
        [CurrentEvents isEmpty] whileTrue: [self readWinQueue].
        int := Process enableInterrupts: false.
        e := CurrentEvents removeFirst.
        Process enableInterrupts: int].! 
imagePalette
    "Returns the current image palette."

    | palette dcRC |
    palette := self class structureNamed: 'TW_PALETTE8'.
    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: self currentDataSource asParameter
        dg: DgImage
        dat: DatPalette8
        msg: MsgGet
        pData: palette asParameter.

    dcRC = TwrcSuccess ifTrue: [ ^palette ].
    "self explainError: 'Error in #imagePalette' number: dcRC."
    ^nil!
imageInfo
    "Returns the current image information."

    | info dcRC |
    info := self class structureNamed: 'TW_IMAGEINFO'.
    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: self currentDataSource asParameter
        dg: DgImage
        dat: DatImageinfo
        msg: MsgGet
        pData: info asParameter.

    dcRC = TwrcSuccess ifTrue: [ ^info ].
    "self explainError: 'Error in #imageInfo' number: dcRC."
    ^nil!
extractOneValue: dcCap
    "Private - Returns one result value from a dcCap structure..."

    | oneValue result itemOffset |
    oneValue := self class
		structureNamed: 'TW_ONEVALUE'
		atAddress: (self globalLock: dcCap hContainer).
    itemOffset := (oneValue definition at: #item) first.

    (oneValue itemType = TwtyInt8 or: [oneValue itemType = TwtyUint8])
    ifTrue: [ result := oneValue byteAtOffset: itemOffset ].

    oneValue itemType = TwtyInt16
    ifTrue: [ result := oneValue shortAtOffset: itemOffset ].

    oneValue itemType = TwtyUint16
    ifTrue: [ result := oneValue uShortAtOffset: itemOffset ].

    oneValue itemType = TwtyInt32
    ifTrue: [ result := oneValue longAtOffset: itemOffset ].

    oneValue itemType = TwtyUint32
    ifTrue: [ result := oneValue uLongAtOffset: itemOffset ].

    oneValue itemType = TwtyBool
    ifTrue: [ result := oneValue booleanAtOffset: itemOffset ].

    self globalUnlock: dcCap hContainer.
    ^result!   
currentUIParent
    "Private - Returns the uiParent window in use..."

    (self isValidWindow: parentUI) ifTrue: [ ^parentUI ].

    parentUI isNil ifTrue: [ parentUI := self activeMainWindow ].
    ^parentUI!  
getCompleteImageFile
    "Private - Ask the Source to start sending the pending image file transfer mode..."

    | dcRC setup result |
	result := TwainFileResult for: self.
    setup := self class structureNamed: 'TW_SETUPFILEXFER'.
	setup
		fileName: result fileName asParameter;
		format: result format;
		vRefNum: 0;
		yourself.
    self fileTransferProperties: setup.
	(File newFile: result fileName) close.
    dcRC := self
		dsmEntry: self appIdentity asParameter
		pDest: self currentDataSource asParameter
		dg: DgImage
		dat: DatImagefilexfer
		msg: MsgGet
		pData: nil asParameter.
    (dcRC = TwrcXferdone or: [ dcRC = TwrcSuccess ])
	ifTrue: [ self acquired: result ]
	ifFalse: [ ^nil ].
	^result! 
openDSM
    "Private - Open the Data Source Manager.
    Returns : current state of the Source Manager"

    | dcRC |
    self isDSMOpen ifTrue: [ ^dsmOpen ].

    dcRC := self
        dsmEntry: self appIdentity asParameter
        pDest: nil
        dg: DgControl
        dat: DatParent
        msg: MsgOpendsm
        pData: self currentUIParent asParameter.

    dsmOpen := dcRC = TwccSuccess.

    dsmOpen ifFalse: [
        self explainError: 'Error when opening Data Source Manager' number: dcRC.
    ].

    ^dsmOpen! 
for: aScanner
	" Returns an instance of the receiver working on aScanner. "

	^self new
		scanner: aScanner;
		yourself! 
comment
    " Documentation - Returns the receiver's comment string. "

    ^'Implements support for image chunks returned during scanning.'!  
fileName
    "Private - Answer the receiver's file name."
     ^'Twain_32.Dll'
!   
countries
    "Private - Returns the country table..."

  ^#(
    'Argentina' 54
    'Usa' 1
    'Afghanistan' 1001
    'Algeria' 213
    'Americansamoa' 684
    'Andorra' 33
    'Angola' 1002
    'Anguilla' 8090
    'Antigua' 8091
    'Argentina' 54
    'Aruba' 297
    'Ascensioni' 247
    'Australia' 61
    'Austria' 43
    'Bahamas' 8092
    'Bahrain' 973
    'Bangladesh' 880
    'Barbados' 8093
    'Belgium' 32
    'Belize' 501
    'Benin' 229
    'Bermuda' 8094
    'Bhutan' 1003
    'Bolivia' 591
    'Botswana' 267
    'Brazil' 55
    'Britain' 6
    'Britvirginis' 8095
    'Brunei' 673
    'Bulgaria' 359
    'Burkinafaso' 1004
    'Burma' 1005
    'Burundi' 1006
    'Camaroon' 237
    'Canada' 2
    'Capeverdeis' 238
    'Caymanis' 8096
    'Centralafrep' 1007
    'Chad' 1008
    'Chile' 56
    'China' 86
    'Christmasis' 1009
    'Cocosis' 1009
    'Colombia' 57
    'Comoros' 1010
    'Congo' 1011
    'Cookis' 1012
    'CostaRica' 506
    'Cuba' 5
    'Cyprus' 357
    'Czechoslovakia' 42
    'Denmark' 45
    'Djibouti' 1013
    'Domincanrep' 8098
    'Dominica' 8097
    'Easteris' 1014
    'Ecuador' 593
    'Egypt' 20
    'Elsalvador' 503
    'Eqguinea' 1015
    'Ethiopia' 251
    'Faeroeis' 298
    'Falklandis' 1016
    'Fijiislands' 679
    'Finland' 358
    'France' 33
    'Frantilles' 596
    'Frguiana' 594
    'Frpolyneisa' 689
    'Futanais' 1043
    'Gabon' 241
    'Gambia' 220
    'Germany' 49
    'Ghana' 233
    'Gibralter' 350
    'Greece' 30
    'Greenland' 299
    'Grenada' 8099
    'Grenedines' 8015
    'Guadeloupe' 590
    'Guam' 671
    'Guantanamobay' 5399
    'Guatemala' 502
    'Guinea' 224
    'Guineabissau' 1017
    'Guyana' 592
    'Haiti' 509
    'Honduras' 504
    'Hong Kong' 852
    'Hungary' 36
    'Iceland' 354
    'India' 91
    'Indonesia' 62
    'Iran' 98
    'Iraq' 964
    'Ireland' 353
    'Israel' 972
    'Italy' 39
    'Ivory Coast' 225
    'Jamaica' 8010
    'Japan' 81
    'Jordan' 962
    'Kenya' 254
    'Kiribati' 1018
    'Korea' 82
    'Kuwait' 965
    'Laos' 1019
    'Lebanon' 1020
    'Liberia' 231
    'Libya' 218
    'Liechtenstein' 41
    'Luxenbourg' 352
    'Macao' 853
    'Madagascar' 1021
    'Malawi' 265
    'Malaysia' 60
    'Maldives' 960
    'Mali' 1022
    'Malta' 356
    'Marshallis' 692
    'Mauritania' 1023
    'Mauritius' 230
    'Mexico' 3
    'Micronesia' 691
    'Miquelon' 508
    'Monaco' 33
    'Mongolia' 1024
    'Montserrat' 8011
    'Morocco' 212
    'Mozambique' 1025
    'Namibia' 264
    'Nauru' 1026
    'Nepal' 977
    'Nethantilles' 599
    'Netherlands' 31
    'Nevis' 8012
    'Newcaledonia' 687
    'Newzealand' 64
    'Nicaragua' 505
    'Niger' 227
    'Nigeria' 234
    'Niue' 1027
    'Norfolki' 1028
    'Norway' 47
    'Oman' 968
    'Pakistan' 92
    'Palau' 1029
    'Panama' 507
    'Paraguay' 595
    'Peru' 51
    'Phillippines' 63
    'Pitcairnis' 1030
    'Pnewguinea' 675
    'Poland' 48
    'Portugal' 351
    'Qatar' 974
    'Reunioni' 1031
    'Romania' 40
    'Rwanda' 250
    'Saipan' 670
    'Sanmarino' 39
    'Saotome' 1033
    'Saudiarabia' 966
    'Senegal' 221
    'Seychellesis' 1034
    'Sierraleone' 1035
    'Singapore' 65
    'Solomonis' 1036
    'Somali' 1037
    'SouthAfrica' 27
    'Spain' 34
    'Srilanka' 94
    'Sthelena' 1032
    'Stkitts' 8013
    'Stlucia' 8014
    'Stpierre' 508
    'Stvincent' 8015
    'Sudan' 1038
    'Suriname' 597
    'Swaziland' 268
    'Sweden' 46
    'Switzerland' 41
    'Syria' 1039
    'Taiwan' 886
    'Tanzania' 255
    'Thailand' 66
    'Tobago' 8016
    'Togo' 228
    'Tongais' 676
    'Trinidad' 8016
    'Tunisia' 216
    'Turkey' 90
    'Turkscaicos' 8017
    'Tuvalu' 1040
    'Uaemirates' 971
    'Uganda' 256
    'Unitedkingdom' 44
    'Uruguay' 598
    'Usa' 1
    'Ussr' 7
    'Vanuatu' 1041
    'Vaticancity' 39
    'Venezuela' 58
    'Wake' 1042
    'Wallisis' 1043
    'Westernsahara' 1044
    'Westernsamoa' 1045
    'Yemen' 1046
    'Yugoslavia' 38
    'Zaire' 243
    'Zambia' 260
    'Zimbabwe' 263
    )!
initializeConstants
    "Private - Initializes the constants for the interface..."

  TwainConstants := Dictionary new.

  TwainConstants
    at: 'CapAuthor' put: 4096;
    at: 'CapAutofeed' put: 4103;
    at: 'CapCaption' put: 4097;
    at: 'CapClearpage' put: 4104;
    at: 'CapCustombase' put: 32768;
    at: 'CapDeviceonline' put: 4109;
    at: 'CapExtendedcaps' put: 4102;
    at: 'CapFeederenabled' put: 4098;
    at: 'CapFeederloaded' put: 4099;
    at: 'CapFeedpage' put: 4105;
    at: 'CapIndicators' put: 4107;
    at: 'CapPaperdetectable' put: 4109;
    at: 'CapRewindpage' put: 4106;
    at: 'CapSupportedcaps' put: 4101;
    at: 'CapSupportedcapsext' put: 4108;
    at: 'CapTimedate' put: 4100;
    at: 'CapUicontrollable' put: 4110;
    at: 'CapXfercount' put: 1;
    at: 'DatCapability' put: 1;
    at: 'DatCiecolor' put: 262;
    at: 'DatCustombase' put: 32768;
    at: 'DatEvent' put: 2;
    at: 'DatGrayresponse' put: 263;
    at: 'DatIdentity' put: 3;
    at: 'DatImagefilexfer' put: 261;
    at: 'DatImageinfo' put: 257;
    at: 'DatImagelayout' put: 258;
    at: 'DatImagememxfer' put: 259;
    at: 'DatImagenativexfer' put: 260;
    at: 'DatJpegcompression' put: 265;
    at: 'DatNull' put: 0;
    at: 'DatPalette8' put: 266;
    at: 'DatParent' put: 4;
    at: 'DatPendingxfers' put: 5;
    at: 'DatRgbresponse' put: 264;
    at: 'DatSetupfilexfer' put: 7;
    at: 'DatSetupmemxfer' put: 6;
    at: 'DatStatus' put: 8;
    at: 'DatTwunkidentity' put: 11;
    at: 'DatUserinterface' put: 9;
    at: 'DatXfergroup' put: 10;
    at: 'DgControl' put: 1;
    at: 'DgImage' put: 2;
    at: 'IcapAutobright' put: 4352;
    at: 'IcapBitdepth' put: 4395;
    at: 'IcapBitdepthreduction' put: 4396;
    at: 'IcapBitorder' put: 4380;
    at: 'IcapBitordercodes' put: 4390;
    at: 'IcapBrightness' put: 4353;
    at: 'IcapCcittkfactor' put: 4381;
    at: 'IcapCompression' put: 256;
    at: 'IcapContrast' put: 4355;
    at: 'IcapCusthalftone' put: 4356;
    at: 'IcapExposuretime' put: 4357;
    at: 'IcapFilter' put: 4358;
    at: 'IcapFlashused' put: 4359;
    at: 'IcapFrames' put: 4372;
    at: 'IcapGamma' put: 4360;
    at: 'IcapHalftones' put: 4361;
    at: 'IcapHighlight' put: 4362;
    at: 'IcapImagefileformat' put: 4364;
    at: 'IcapJpegpixeltype' put: 4392;
    at: 'IcapLampstate' put: 4365;
    at: 'IcapLightpath' put: 4382;
    at: 'IcapLightsource' put: 4366;
    at: 'IcapMaxframes' put: 4378;
    at: 'IcapOrientation' put: 4368;
    at: 'IcapPhysicalheight' put: 4370;
    at: 'IcapPhysicalwidth' put: 4369;
    at: 'IcapPixelflavor' put: 4383;
    at: 'IcapPixelflavorcodes' put: 4391;
    at: 'IcapPixeltype' put: 257;
    at: 'IcapPlanarchunky' put: 4384;
    at: 'IcapRotation' put: 4385;
    at: 'IcapShadow' put: 4371;
    at: 'IcapSupportedsizes' put: 4386;
    at: 'IcapThreshold' put: 4387;
    at: 'IcapTiles' put: 4379;
    at: 'IcapTimefill' put: 4394;
    at: 'IcapUndefinedimagesize' put: 4397;
    at: 'IcapUnits' put: 258;
    at: 'IcapXfermech' put: 259;
    at: 'IcapXnativeresolution' put: 4374;
    at: 'IcapXresolution' put: 4376;
    at: 'IcapXscaling' put: 4388;
    at: 'IcapYnativeresolution' put: 4375;
    at: 'IcapYresolution' put: 4377;
    at: 'IcapYscaling' put: 4389;
    at: 'MsgCheckstatus' put: 513;
    at: 'MsgCloseds' put: 1026;
    at: 'MsgClosedsm' put: 770;
    at: 'MsgClosedsreq' put: 258;
    at: 'MsgCustombase' put: 32768;
    at: 'MsgDisableds' put: 1281;
    at: 'MsgEnableds' put: 1282;
    at: 'MsgEndxfer' put: 1793;
    at: 'MsgGet' put: 1;
    at: 'MsgGetcurrent' put: 2;
    at: 'MsgGetdefault' put: 3;
    at: 'MsgGetfirst' put: 4;
    at: 'MsgGetnext' put: 5;
    at: 'MsgNull' put: 0;
    at: 'MsgOpends' put: 1025;
    at: 'MsgOpendsm' put: 769;
    at: 'MsgProcessevent' put: 1537;
    at: 'MsgReset' put: 7;
    at: 'MsgSet' put: 6;
    at: 'MsgUserselect' put: 1027;
    at: 'MsgXferready' put: 257;
    at: 'TwboLsbfirst' put: 0;
    at: 'TwboMsbfirst' put: 1;
    at: 'TwbrCusthalftone' put: 2;
    at: 'TwbrDiffusion' put: 3;
    at: 'TwbrHalftone' put: 1;
    at: 'TwbrThreshold' put: 0;
    at: 'TwccBadcap' put: 6;
    at: 'TwccBaddest' put: 12;
    at: 'TwccBadprotocol' put: 9;
    at: 'TwccBadvalue' put: 10;
    at: 'TwccBummer' put: 1;
    at: 'TwccCustombase' put: 32768;
    at: 'TwccLowmemory' put: 2;
    at: 'TwccMaxconnections' put: 4;
    at: 'TwccNods' put: 3;
    at: 'TwccOperationerror' put: 5;
    at: 'TwccSeqerror' put: 11;
    at: 'TwccSuccess' put: 0;
    at: 'TwcpGroup31d' put: 2;
    at: 'TwcpGroup31deol' put: 3;
    at: 'TwcpGroup32d' put: 4;
    at: 'TwcpGroup4' put: 5;
    at: 'TwcpJpeg' put: 6;
    at: 'TwcpLzw' put: 7;
    at: 'TwcpNone' put: 0;
    at: 'TwcpPackbits' put: 1;
    at: 'TwffBmp' put: 2;
    at: 'TwffJfif' put: 4;
    at: 'TwffPict' put: 1;
    at: 'TwffTiff' put: 0;
    at: 'TwffXbm' put: 3;
    at: 'TwftBlack' put: 8;
    at: 'TwftBlue' put: 2;
    at: 'TwftCyan' put: 5;
    at: 'TwftGreen' put: 1;
    at: 'TwftMagenta' put: 6;
    at: 'TwftNone' put: 3;
    at: 'TwftRed' put: 0;
    at: 'TwftWhite' put: 4;
    at: 'TwftYellow' put: 7;
    at: 'TwlpReflective' put: 0;
    at: 'TwlpTransmissive' put: 1;
    at: 'TwlsBlue' put: 2;
    at: 'TwlsGreen' put: 1;
    at: 'TwlsIr' put: 6;
    at: 'TwlsNone' put: 3;
    at: 'TwlsRed' put: 0;
    at: 'TwlsUv' put: 5;
    at: 'TwlsWhite' put: 4;
    at: 'TwmfAppowns' put: 1;
    at: 'TwmfDsmowns' put: 2;
    at: 'TwmfDsowns' put: 4;
    at: 'TwmfHandle' put: 16;
    at: 'TwmfPointer' put: 8;
    at: 'TwonArray' put: 3;
    at: 'TwonDontcare16' put: 65535;
    at: 'TwonDontcare32' put: 4294967295;
    at: 'TwonDontcare8' put: 255;
    at: 'TwonDsmcodeid' put: 63;
    at: 'TwonDsmid' put: 461;
    at: 'TwonEnumeration' put: 4;
    at: 'TwonIconid' put: 962;
    at: 'TwonOnevalue' put: 5;
    at: 'TwonProtocolmajor' put: 1;
    at: 'TwonProtocolminor' put: 5;
    at: 'TwonRange' put: 6;
    at: 'TworLandscape' put: 0;
    at: 'TworPortrait' put: 0;
    at: 'TworRot0' put: 0;
    at: 'TworRot180' put: 2;
    at: 'TworRot270' put: 3;
    at: 'TworRot90' put: 1;
    at: 'TwpaCmy' put: 2;
    at: 'TwpaGray' put: 1;
    at: 'TwpaRgb' put: 0;
    at: 'TwpcChunky' put: 0;
    at: 'TwpcPlanar' put: 1;
    at: 'TwpfChocolate' put: 0;
    at: 'TwpfVanilla' put: 1;
    at: 'TwptBw' put: 0;
    at: 'TwptCiexyz' put: 8;
    at: 'TwptCmy' put: 4;
    at: 'TwptCmyk' put: 5;
    at: 'TwptGray' put: 1;
    at: 'TwptPalette' put: 3;
    at: 'TwptRgb' put: 2;
    at: 'TwptYuv' put: 6;
    at: 'TwptYuvk' put: 7;
    at: 'TwrcCancel' put: 3;
    at: 'TwrcCheckstatus' put: 2;
    at: 'TwrcCustombase' put: 32768;
    at: 'TwrcDsevent' put: 4;
    at: 'TwrcEndoflist' put: 7;
    at: 'TwrcFailure' put: 1;
    at: 'TwrcNotdsevent' put: 5;
    at: 'TwrcSuccess' put: 0;
    at: 'TwrcXferdone' put: 6;
    at: 'TwssA4letter' put: 1;
    at: 'TwssA5' put: 5;
    at: 'TwssB' put: 8;
    at: 'TwssB4' put: 6;
    at: 'TwssB5letter' put: 2;
    at: 'TwssB6' put: 7;
    at: 'TwssNone' put: 0;
    at: 'TwssUslegal' put: 4;
    at: 'TwssUsletter' put: 3;
    at: 'TwsxFile' put: 1;
    at: 'TwsxMemory' put: 2;
    at: 'TwsxNative' put: 0;
    at: 'TwtyBool' put: 6;
    at: 'TwtyFix32' put: 7;
    at: 'TwtyFrame' put: 8;
    at: 'TwtyInt16' put: 1;
    at: 'TwtyInt32' put: 2;
    at: 'TwtyInt8' put: 0;
    at: 'TwtyStr128' put: 11;
    at: 'TwtyStr255' put: 12;
    at: 'TwtyStr32' put: 9;
    at: 'TwtyStr64' put: 10;
    at: 'TwtyUint16' put: 4;
    at: 'TwtyUint32' put: 5;
    at: 'TwtyUint8' put: 3;
    at: 'TwunCentimeters' put: 1;
    at: 'TwunInches' put: 0;
    at: 'TwunPicas' put: 2;
    at: 'TwunPixels' put: 5;
    at: 'TwunPoints' put: 3;
    at: 'TwunTwips' put: 4;
    at: 'TwcyAfghanistan' put: 1001;
    at: 'TwcyAlgeria' put: 213;
    at: 'TwcyAmericansamoa' put: 684;
    at: 'TwcyAndorra' put: 33;
    at: 'TwcyAngola' put: 1002;
    at: 'TwcyAnguilla' put: 8090;
    at: 'TwcyAntigua' put: 8091;
    at: 'TwcyArgentina' put: 54;
    at: 'TwcyAruba' put: 297;
    at: 'TwcyAscensioni' put: 247;
    at: 'TwcyAustralia' put: 61;
    at: 'TwcyAustria' put: 43;
    at: 'TwcyBahamas' put: 8092;
    at: 'TwcyBahrain' put: 973;
    at: 'TwcyBangladesh' put: 880;
    at: 'TwcyBarbados' put: 8093;
    at: 'TwcyBelgium' put: 32;
    at: 'TwcyBelize' put: 501;
    at: 'TwcyBenin' put: 229;
    at: 'TwcyBermuda' put: 8094;
    at: 'TwcyBhutan' put: 1003;
    at: 'TwcyBolivia' put: 591;
    at: 'TwcyBotswana' put: 267;
    at: 'TwcyBrazil' put: 55;
    at: 'TwcyBritain' put: 6;
    at: 'TwcyBritvirginis' put: 8095;
    at: 'TwcyBrunei' put: 673;
    at: 'TwcyBulgaria' put: 359;
    at: 'TwcyBurkinafaso' put: 1004;
    at: 'TwcyBurma' put: 1005;
    at: 'TwcyBurundi' put: 1006;
    at: 'TwcyCamaroon' put: 237;
    at: 'TwcyCanada' put: 2;
    at: 'TwcyCapeverdeis' put: 238;
    at: 'TwcyCaymanis' put: 8096;
    at: 'TwcyCentralafrep' put: 1007;
    at: 'TwcyChad' put: 1008;
    at: 'TwcyChile' put: 56;
    at: 'TwcyChina' put: 86;
    at: 'TwcyChristmasis' put: 1009;
    at: 'TwcyCocosis' put: 1009;
    at: 'TwcyColombia' put: 57;
    at: 'TwcyComoros' put: 1010;
    at: 'TwcyCongo' put: 1011;
    at: 'TwcyCookis' put: 1012;
    at: 'TwcyCostarica' put: 506;
    at: 'TwcyCuba' put: 5;
    at: 'TwcyCyprus' put: 357;
    at: 'TwcyCzechoslovakia' put: 42;
    at: 'TwcyDenmark' put: 45;
    at: 'TwcyDjibouti' put: 1013;
    at: 'TwcyDomincanrep' put: 8098;
    at: 'TwcyDominica' put: 8097;
    at: 'TwcyEasteris' put: 1014;
    at: 'TwcyEcuador' put: 593;
    at: 'TwcyEgypt' put: 20;
    at: 'TwcyElsalvador' put: 503;
    at: 'TwcyEqguinea' put: 1015;
    at: 'TwcyEthiopia' put: 251;
    at: 'TwcyFaeroeis' put: 298;
    at: 'TwcyFalklandis' put: 1016;
    at: 'TwcyFijiislands' put: 679;
    at: 'TwcyFinland' put: 358;
    at: 'TwcyFrance' put: 33;
    at: 'TwcyFrantilles' put: 596;
    at: 'TwcyFrguiana' put: 594;
    at: 'TwcyFrpolyneisa' put: 689;
    at: 'TwcyFutanais' put: 1043;
    at: 'TwcyGabon' put: 241;
    at: 'TwcyGambia' put: 220;
    at: 'TwcyGermany' put: 49;
    at: 'TwcyGhana' put: 233;
    at: 'TwcyGibralter' put: 350;
    at: 'TwcyGreece' put: 30;
    at: 'TwcyGreenland' put: 299;
    at: 'TwcyGrenada' put: 8099;
    at: 'TwcyGrenedines' put: 8015;
    at: 'TwcyGuadeloupe' put: 590;
    at: 'TwcyGuam' put: 671;
    at: 'TwcyGuantanamobay' put: 5399;
    at: 'TwcyGuatemala' put: 502;
    at: 'TwcyGuinea' put: 224;
    at: 'TwcyGuineabissau' put: 1017;
    at: 'TwcyGuyana' put: 592;
    at: 'TwcyHaiti' put: 509;
    at: 'TwcyHonduras' put: 504;
    at: 'TwcyHongkong' put: 852;
    at: 'TwcyHungary' put: 36;
    at: 'TwcyIceland' put: 354;
    at: 'TwcyIndia' put: 91;
    at: 'TwcyIndonesia' put: 62;
    at: 'TwcyIran' put: 98;
    at: 'TwcyIraq' put: 964;
    at: 'TwcyIreland' put: 353;
    at: 'TwcyIsrael' put: 972;
    at: 'TwcyItaly' put: 39;
    at: 'TwcyIvorycoast' put: 225;
    at: 'TwcyJamaica' put: 8010;
    at: 'TwcyJapan' put: 81;
    at: 'TwcyJordan' put: 962;
    at: 'TwcyKenya' put: 254;
    at: 'TwcyKiribati' put: 1018;
    at: 'TwcyKorea' put: 82;
    at: 'TwcyKuwait' put: 965;
    at: 'TwcyLaos' put: 1019;
    at: 'TwcyLebanon' put: 1020;
    at: 'TwcyLiberia' put: 231;
    at: 'TwcyLibya' put: 218;
    at: 'TwcyLiechtenstein' put: 41;
    at: 'TwcyLuxenbourg' put: 352;
    at: 'TwcyMacao' put: 853;
    at: 'TwcyMadagascar' put: 1021;
    at: 'TwcyMalawi' put: 265;
    at: 'TwcyMalaysia' put: 60;
    at: 'TwcyMaldives' put: 960;
    at: 'TwcyMali' put: 1022;
    at: 'TwcyMalta' put: 356;
    at: 'TwcyMarshallis' put: 692;
    at: 'TwcyMauritania' put: 1023;
    at: 'TwcyMauritius' put: 230;
    at: 'TwcyMexico' put: 3;
    at: 'TwcyMicronesia' put: 691;
    at: 'TwcyMiquelon' put: 508;
    at: 'TwcyMonaco' put: 33;
    at: 'TwcyMongolia' put: 1024;
    at: 'TwcyMontserrat' put: 8011;
    at: 'TwcyMorocco' put: 212;
    at: 'TwcyMozambique' put: 1025;
    at: 'TwcyNamibia' put: 264;
    at: 'TwcyNauru' put: 1026;
    at: 'TwcyNepal' put: 977;
    at: 'TwcyNethantilles' put: 599;
    at: 'TwcyNetherlands' put: 31;
    at: 'TwcyNevis' put: 8012;
    at: 'TwcyNewcaledonia' put: 687;
    at: 'TwcyNewzealand' put: 64;
    at: 'TwcyNicaragua' put: 505;
    at: 'TwcyNiger' put: 227;
    at: 'TwcyNigeria' put: 234;
    at: 'TwcyNiue' put: 1027;
    at: 'TwcyNorfolki' put: 1028;
    at: 'TwcyNorway' put: 47;
    at: 'TwcyOman' put: 968;
    at: 'TwcyPakistan' put: 92;
    at: 'TwcyPalau' put: 1029;
    at: 'TwcyPanama' put: 507;
    at: 'TwcyParaguay' put: 595;
    at: 'TwcyPeru' put: 51;
    at: 'TwcyPhillippines' put: 63;
    at: 'TwcyPitcairnis' put: 1030;
    at: 'TwcyPnewguinea' put: 675;
    at: 'TwcyPoland' put: 48;
    at: 'TwcyPortugal' put: 351;
    at: 'TwcyQatar' put: 974;
    at: 'TwcyReunioni' put: 1031;
    at: 'TwcyRomania' put: 40;
    at: 'TwcyRwanda' put: 250;
    at: 'TwcySaipan' put: 670;
    at: 'TwcySanmarino' put: 39;
    at: 'TwcySaotome' put: 1033;
    at: 'TwcySaudiarabia' put: 966;
    at: 'TwcySenegal' put: 221;
    at: 'TwcySeychellesis' put: 1034;
    at: 'TwcySierraleone' put: 1035;
    at: 'TwcySingapore' put: 65;
    at: 'TwcySolomonis' put: 1036;
    at: 'TwcySomali' put: 1037;
    at: 'TwcySouthafrica' put: 27;
    at: 'TwcySpain' put: 34;
    at: 'TwcySrilanka' put: 94;
    at: 'TwcySthelena' put: 1032;
    at: 'TwcyStkitts' put: 8013;
    at: 'TwcyStlucia' put: 8014;
    at: 'TwcyStpierre' put: 508;
    at: 'TwcyStvincent' put: 8015;
    at: 'TwcySudan' put: 1038;
    at: 'TwcySuriname' put: 597;
    at: 'TwcySwaziland' put: 268;
    at: 'TwcySweden' put: 46;
    at: 'TwcySwitzerland' put: 41;
    at: 'TwcySyria' put: 1039;
    at: 'TwcyTaiwan' put: 886;
    at: 'TwcyTanzania' put: 255;
    at: 'TwcyThailand' put: 66;
    at: 'TwcyTobago' put: 8016;
    at: 'TwcyTogo' put: 228;
    at: 'TwcyTongais' put: 676;
    at: 'TwcyTrinidad' put: 8016;
    at: 'TwcyTunisia' put: 216;
    at: 'TwcyTurkey' put: 90;
    at: 'TwcyTurkscaicos' put: 8017;
    at: 'TwcyTuvalu' put: 1040;
    at: 'TwcyUaemirates' put: 971;
    at: 'TwcyUganda' put: 256;
    at: 'TwcyUnitedkingdom' put: 44;
    at: 'TwcyUruguay' put: 598;
    at: 'TwcyUsa' put: 1;
    at: 'TwcyUssr' put: 7;
    at: 'TwcyVanuatu' put: 1041;
    at: 'TwcyVaticancity' put: 39;
    at: 'TwcyVenezuela' put: 58;
    at: 'TwcyWake' put: 1042;
    at: 'TwcyWallisis' put: 1043;
    at: 'TwcyWesternsahara' put: 1044;
    at: 'TwcyWesternsamoa' put: 1045;
    at: 'TwcyYemen' put: 1046;
    at: 'TwcyYugoslavia' put: 38;
    at: 'TwcyZaire' put: 243;
    at: 'TwcyZambia' put: 260;
    at: 'TwcyZimbabwe' put: 263;
    at: 'TwlgDan' put: 0;
    at: 'TwlgDut' put: 1;
    at: 'TwlgEng' put: 2;
    at: 'TwlgFcf' put: 3;
    at: 'TwlgFin' put: 4;
    at: 'TwlgFrn' put: 5;
    at: 'TwlgGer' put: 6;
    at: 'TwlgIce' put: 7;
    at: 'TwlgItn' put: 8;
    at: 'TwlgNor' put: 9;
    at: 'TwlgPor' put: 10;
    at: 'TwlgSpa' put: 11;
    at: 'TwlgSwe' put: 12;
    at: 'TwlgUsa' put: 13;
    yourself.!   
initializeStructures
    "Private - Initializes the structures for the interface...
        Twain32DLL initializeStructures
    "
    self
        registerType: 'TW_VERSION'
        as: #(
            (majorNum word) "Major revision number of the software."
            (minorNum word) "Incremental revision number of the software."
            (language word) "DCLG_xx"
            (country word) "DCCY_xx"
            (info 34 asCstring) "1.0b3 Beta release"
        );
        registerType: 'TW_IDENTITY' "Identifies the program/library/code resource."
        as: #(
            (id ulong) "Unique number.  In Windows, client hWnd "
            (version 'TW_VERSION') "Identifies the piece of code"
            (protocolMajor word) (protocolMinor word)
            (supportedGroups ulong) "Bit field OR combination of DG_ constants"
            (manufacturer 34 asCstring) "Manufacturer name"
            (productFamily 34 asCstring) "Product family name"
            (productName 34 asCstring) "Product name"
        );
        registerType: 'TW_STATUS' "Gets detailed status info from a data source with this."
        as: #(
            (code ushort) "Any TWCC_ constant"
            (reserved ushort) "Future expansion space"
        );
        registerType: 'TW_USERINTERFACE' "Coordinates UI with data source."
        as: #(
            (showUI ushort) "TRUE if DS should bring up its UI"
            (modalUI ushort) "For Mac only - true if the DS's UI is modal"
            (hParent ulong) "For windows only - App window handle"
        );
        registerType: 'TW_EVENT' "For passing events down to the DS."
        as: #(
            (pEvent 4) "Windows pMSG or Mac pEvent."
            (dcMessage ushort) "TW msg from data source, e.g. MSG_XFERREADY"
        );
        registerType: 'TW_PENDINGXFERS' "Used with MSG_ENDXFER to indicate additional data."
        as: #(
            (count ushort) "Number of additional 'images' pending."
            (reserved ulong)
        );
        registerType: 'TW_IMAGEINFO'
        as: #(
            (xResolution long) "Resolution in the horizontal"
            (yResolution long) "Resolution in the vertical"
            (imageWidth long) "Columns in the image, -1 if unknown by DS"
            (imageLength long) "Rows in the image, -1 if unknown by DS"
            (samplesPerPixel short) "Number of samples per pixel, 3 for RGB"
            (bitsPerSample 16 asString) "Number of bits for each sample"
            (bitsPerPixel short) "Number of bits for each padded pixel"
            (planar ushort) "True if Planar, False if chunky"
            (pixelType short) "How to interp data; photo interp (DCPT_)"
            (compression ushort) "How the data is compressed (DCCP_xxxx)"
        );
        "Defines a frame rectangle in ICAP_UNITS coordinates."
        registerType: 'TW_FRAME'
        as: #(
            (top long)
            (left long)
            (right long)
            (bottom long)
        );
        registerType: 'TW_IMAGELAYOUT' "Provides image layout information in current units."
        as: #(
            (frame 'TW_FRAME') "Frame coords within larger document"
            (documentNumber ulong)
            (pageNumber ulong) "Reset when you go to next document "
            (frameNumber ulong) "Reset when you go to next page"
        );
        registerType: 'TW_CAPABILITY'
        as: #(
            (cap word) "id of capability to set or get, e.g. CAP_BRIGHTNESS."
            (conType word) "TWON_ONEVALUE, _RANGE, _ENUMERATION or _ARRAY"
            (hContainer ulong) "Handle to container of type Dat_xxx"
        );
        registerType: 'TW_ONEVALUE' "Container for one value."
        as: #(
            (itemType word)
            (item ulong)
        );
        registerType: 'TW_RANGE'    "Container for a range of values."
        as: #(
            (itemType ushort)
            (minValue ushort) "Starting value in the range."
            (maxValue ushort) "Final value in the range."
            (stepSize ushort) "Increment from MinValue to MaxValue."
            (defaultValue ushort) "Power-up value."
            (currentValue ushort) "The value that is currently in effect."
        );
        registerType: 'TW_ENUMERATION'  "Container for a collection of values."
        as: #(
            (itemType ushort)
            (numItems ulong) "How many items in ItemList"
            (currentIndex ulong) "Current value is in ItemList[CurrentIndex]"
            (defaultIndex ulong) "Powerup value is in ItemList[DefaultIndex]"
            (itemList (0 byte)) "Array of ItemType values starts here"
        );
        registerType: 'TW_ARRAY' "Container for arrayed values."
        as: #(
            (itemType word)
            (numItems ulong) "How many items in ItemList"
            (itemList (0 byte)) "Array of ItemType values starts here"
        );
        registerType: 'TW_FIX32' "Fixed point structure type."
        as: #(
            (whole short) "maintains the sign"
            (frac word)
        );
        registerType: 'TW_DECODEFUNCTION'
        as: #(
            (startIn long) (breakIn long) (endIn long)
            (startOut long) (breakOut long) (endOut long)
            (gamma long)
            (sampleCount long) "if =0 use the gamma"
        );
        registerType: 'TW_CIEPOINT'
        as: #(
            (x long) (y long) (z long)
        );
        registerType: 'TW_TRANSFORMSTAGE'
        as: #(
            (decode (3 'TW_DECODEFUNCTION'))
            (mix (9 long))
        );
        registerType: 'TW_CIECOLOR'
        as: #(
            (colorSpace word)
            (lowEndian short)
            (veviceDependent short)
            (versionNumber long)
            (stageABC 'TW_TRANSFORMSTAGE')
            (stageLMN 'TW_TRANSFORMSTAGE')
            (whitePoint 'TW_CIEPOINT')
            (blackPoint 'TW_CIEPOINT')
            (whitePaper 'TW_CIEPOINT')
            (blackInk 'TW_CIEPOINT')
            (samples (0 long))
        );
        registerType: 'TW_ELEMENT8'
        as: #(
            (index byte) "Value used to index into the color table."
            (channel1 byte) "First  tri-stimulus value (e.g Red)"
            (channel2 byte) "First  tri-stimulus value (e.g Green)"
            (channel3 byte) "First  tri-stimulus value (e.g Blue)"
        );
        "Used to manage memory buffers."
        registerType: 'TW_MEMORY'
        as: #(
            (flags ulong) "Any combination of the TWMF_ constants."
            (length ulong) "Number of bytes stored in buffer TheMem."
            (theMem ulong) "Pointer or handle to the allocated memory buffer."
        );
        registerType: 'TW_IMAGEMEMXFER' "Used to pass image data (e.g. in strips) from DS to app."
        as: #(
            (compression word) "How the data is compressed."
            (bytesPerRow ulong) "Number of bytes in a row of data."
            (columns ulong) "How many columns."
            (rows ulong) "How many rows."
            (xOffset ulong) "How far from the side of the image."
            (yOffset ulong) "How far from the top of the image."
            (bytesWritten ulong) "How many bytes written in Memory."
            (memory 'TW_MEMORY') "Mem struct used to pass actual image data."
        );
        registerType: 'TW_JPEGCOMPRESSION' "QuantTable, HuffmanDC, HuffmanAC TW_MEMREF -> TW_MEMORY.  Based on JPEG Draft International Std, ver 10918-1"
        as: #(
            (colorSpace word) "One of the TWPT_xxxx values"
            (subSampling ulong) "Two word 'array' for subsampling values"
            (numComponents word) "Number of color components in image"
            (restartFrequency word) "Frequency of restart marker codes in MDU's"
            (quantMap (4 ushort)) "Mapping of components to QuantTables"
            (quantTable (4 ushort)) "Quantization tables"
            (huffmanMap (4 ushort)) "Mapping of components to Huffman tables"
            (huffmanDC (2 ulong)) "DC Huffman tables"
            (huffmanAC (2 ulong)) "AC Huffman tables"
        );
        registerType: 'TW_SETUPFILEXFER' "Sets up DS to app data transfer via a file."
        as: #(
            (fileName 256 asCstring)
            (format ushort) "Any TWFF_ constant"
            (vRefNum short) "Used for Mac only "
        );
        registerType: 'TW_SETUPMEMXFER' "Sets up DS to app data transfer via a memory buffer."
        as: #(
            (minBufSize ulong)
            (maxBufSize ulong)
            (preferred ulong)
        );
        registerType: 'TW_PALETTE8' "Color palette when TWPT_PALETTE pixels xfer'd in mem buf."
        as: #(
            (numColors word) "Number of colors in the color table."
            (paletteType word) "TWPA_xxxx, specifies type of palette."
            (colors (256 'TW_ELEMENT8')) "Array of palette values starts here."
        );
        registerType: 'TW_GRAYRESPONSE'
        as: #( (response (0 byte)) );
        registerType: 'TW_RGBRESPONSE'
        as: #( (response (0 'TW_ELEMENT8')) );
        registerType: 'TW_TWUNKIDENTITY' "Provides DS identity and 'other' information necessary across thunk link."
        as: #(
            (identity 'TW_IDENTITY') "Identity of data source."
            (dsPath 255 asCstring) "Full path and file name of data source."
        );
        registerType: 'TW_TWUNKDSENTRYPARAMS' "Provides DS_Entry parameters over thunk link."
        as: #(
            (destFlag byte) "TRUE if dest is not NULL"
            (dest 'TW_IDENTITY') "Identity of data source (if used)."
            (dataGroup long) "DSM_Entry dataGroup parameter"
            (dataArgType long) "DSM_Entry dataArgType parameter"
            (message long) "DSM_Entry message parameter"
            (pDataSize long) "Size of pData (0 if NULL)"
            (pData ulong) "Based on implementation specifics, a pData parameter makes no sense in this structure, but data (if provided) will be appended in the data block."
        );
        registerType: 'TW_TWUNKDSENTRYRETURN' "Provides DS_Entry results over thunk link."
        as: #(
            (returnCode word) "Thunker DsEntry return code."
            (conditionCode word) "Thunker DsEntry condition code."
            (pDataSize long) "Size of pData (0 if NULL)"
            (pData ulong) "Based on implementation specifics, a pData parameter makes no sense in this structure, but data (if provided) will be appended in the data block."
        );
        registerType: 'TW_CAPEXT' "TWAIN 1.6 CAP_SUPPORTEDCAPSEXT structure."
        as: #(
            (cap word) " Which CAP/ICAP info is relevant to"
            (properties word) "Messages this CAP/ICAP supports"
        );
      yourself!
countryNamed: aName
    "Private - Returns the id for country named aName..."

    | stream |
    stream := ReadStream on: self countries.
    [stream atEnd] whileFalse: [
        stream next = aName ifTrue: [^stream next].
        stream next "skip id..."
    ].
    ^1 "default"!
dsEntry: pOrigin dg: dg dat: dat msg: msg pData: pData
    "Private - DS_Entry, the entry point provided by a Data Source...
    Parameters:
        pOrigin Identifies the source module of the message.
                    This could identify an application or the Data Source Manager.
        DG      The Data Group. Example: DG_IMAGE.
        DAT     The Data Attribute Type. Example: DAT_IMAGEMEMXFER.
        MSG    The message.  Messages are interpreted by the data source
                    with respect to the Data Group and the Data Attribute Type.
                    Example: MSG_GET.
        pData   A pointer to the data structure or variable identified
                    by the Data Attribute Type.
                    Example: (TW_MEMREF)&ImageMemXfer
                         where ImageMemXfer is a TW_IMAGEMEMXFER structure.
    Returns: ReturnCode  Example: DCRC_SUCCESS.

    Note:
        The DSPROC type is only used by an application when it calls
        a Data Source directly, bypassing the Data Source Manager.

    "
    <api: 'DS_Entry' struct ulong ushort ushort struct ushort>
    ^self invalidArgument! 
dsmEntry: pOrigin pDest: pDest dg: dg dat: dat msg: msg pData: pData
    "Private - DSM_Entry, the only entry point into the Data Source Manager...
    Parameters:
        pOrigin Identifies the source module of the message.
                    This could identify an application, a Source, or the Source Manager.
        pDest   Identifies the destination module for the message.
                    This could identify an application or a data source.
                    If this is NULL, the message goes to the Source Manager.
        DG      The Data Group. Example: DG_IMAGE.
        DAT     The Data Attribute Type. Example: DAT_IMAGEMEMXFER.
        MSG    The message.  Messages are interpreted by the destination module
                    with respect to the Data Group and the Data Attribute Type.
                    Example: MSG_GET.
        pData   A pointer to the data structure or variable identified
                    by the Data Attribute Type.
                    Example: (TW_MEMREF)&ImageMemXfer
                         where ImageMemXfer is a TW_IMAGEMEMXFER structure.
    Returns: ReturnCode  Example: DCRC_SUCCESS.
    "
    <api: 'DSM_Entry' struct struct ulong ushort ushort struct ushort>
    ^self invalidArgument!  
comment
    " Documentation - Returns the receiver's comment string. "

    ^'Native image chunks contains only a single image.
	Native transfers are implemented passing a handle to a memory bitmap.'!  
asDIB
	" Returns the receiver as a DIB. "

	self implementedBySubclass!
displayAt: aPoint with: aPen
	" Display the receiver at aPoint with aPen. "

	^self asDIB displayAt: aPoint with: aPen!
scanner
    "Returns the instance var of the receiver
     for external manipulation"
    ^scanner!
scanner: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    scanner := anObject!   
asDIB
	" Returns the receiver as a DIB. "

	^DIB fromHandle: self hBitmap! 
hBitmap
    "Returns the instance var of the receiver
     for external manipulation"
    ^hBitmap!
hBitmap: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    hBitmap := anObject!   
asDIB
	" Returns the receiver as a DIB. "

	^DIB fromFile: self fileName!  
initializeFileName
    "Returns the instance var of the receiver."

	fileName := self scanner settings fileTransferFileName.!  
fileName
    "Returns the instance var of the receiver."

	fileName isNil ifTrue: [ self initializeFileName ].
    ^fileName! 
format: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    format := anObject! 
initializeFormat
	"Private - Initializes the receiver's format."

	format := self scanner settings fileTransferFormat.!
fileName: anObject
    "Sets the instance var of the receiver
     with anObject. (No checking is done)"
    fileName := anObject! 
format
    "Returns the instance var of the receiver
     for external manipulation"

	format isNil ifTrue: [ self initializeFormat ].
    ^format!  
paletteFor: bitmapInfo
	"Private - Returns palette information for current image."

	| info rgb pixelType pixelFlavor stream palette buildGrayPalette colors |
	info := self scanner imageInfo.
    (bitmapInfo bitCount = 24 or: [ info pixelType = TwptRgb ])
	ifTrue: [
		bitmapInfo clrUsed: 0.
		^nil
	].
	info pixelType = TwptBw
	ifTrue: [
		bitmapInfo clrUsed: 2; clrImportant: 0; yourself.
		rgb := ByteArray new: 4 * 2.
		pixelFlavor := self scanner pixelFlavor.
		(pixelFlavor isNil or: [ pixelFlavor = 0 ])
		ifTrue: [ 5 to: 7 do: [:i| rgb at: i put: 16rFF ]. ]
		ifFalse: [ 1 to: 3 do: [:i| rgb at: i put: 16rFF ]. ].
		^rgb
	].
	buildGrayPalette := [
		bitmapInfo clrUsed: 256; clrImportant: 0; yourself.
		rgb := ByteArray new:  (
            WinBitmapInfo
                colorTableSize: bitmapInfo bitCount
                clrUsed: bitmapInfo clrUsed).
		stream := WriteStream on: rgb.
		0 to: 255 do: [:i|
			stream next: 3 put: i; nextPut: 0.
		].
		^rgb
	].
	info pixelType = TwptGray ifTrue: [ buildGrayPalette value ].

	"info pixelType = TwptPalette TwptCmy TwptCmyk TwptYuv TwptYuvk TwptCiexyz"
	palette := self scanner imagePalette.
	palette isNil ifTrue: [ buildGrayPalette value ].
	bitmapInfo clrUsed: palette numColors; clrImportant: 0; yourself.
	rgb := ByteArray new:  (
		WinBitmapInfo
			colorTableSize: bitmapInfo bitCount
			clrUsed: bitmapInfo clrUsed).
	stream := WriteStream on: rgb.
	1 to: palette numColors do: [:index|
		colors := palette colors at: index.
		stream
			nextPut: colors channel3;
			nextPut: colors channel2;
			nextPut: colors channel1;
			nextPut: 0.
	].
	^rgb!
asDIB
	" Returns the receiver as a DIB. "

	^self dib! 
completed
	"The scanning has been completed."!   
addBytes: count from: aMemoryBlock
	"Add a memory block chunk to current dib."

	| bits stop |
	(aMemoryBlock isNil or: [ self dib isNil ]) ifTrue: [ ^nil ].
	bits := self dib bits.
	stop := offset + count min: bits size.
	bits
		replaceFrom: offset
        to: stop
        with: aMemoryBlock
        startingAt: 1.
	offset := stop.
	^aMemoryBlock!
cancelled
	"The scann has been cancelled perform dib cleanUp."

	dib := nil.!  
dib
    "Private - Returns current image DIB."

    | info bitmapInfo bits bytes rgb |
    dib notNil ifTrue: [ ^dib ].

    "build DIB instance to hold current image
    and set current offset to zero."
    info := self scanner imageInfo.
    bitmapInfo := WinBitmapInfoHeader new.
    bitmapInfo
        width: info imageWidth;
        height: info imageLength;
        planes: 1;
        bitCount: info bitsPerPixel;
        compression: 0"BiRgb";
        yourself.
    rgb := self paletteFor: bitmapInfo.
    bitmapInfo := WinBitmapInfo fromBytes: (
        rgb isNil
            ifTrue: [ bitmapInfo contents ]
            ifFalse: [ bitmapInfo contents, rgb ] ).
    (bitmapInfo respondsTo: #biSize:)
    ifTrue: [ "Visual Smalltalk"
        #platformDependent.
        bitmapInfo biSize: WinBitmapInfoHeader sizeInBytes
    ]
    ifFalse: [
        bitmapInfo size:  WinBitmapInfoHeader sizeInBytes
    ].
    bitmapInfo sizeImage = 0
    ifTrue: [
        bytes := (((bitmapInfo width * bitmapInfo bitCount) + 31) // 32 * 4) * bitmapInfo height
    ]
    ifFalse: [bytes := bitmapInfo sizeImage].

    bits := ByteArray new: bytes.
    offset := 1. "next add will start filling the image."
    dib := DIB new
        bitmapInfo: bitmapInfo;
        bits: bits;
        yourself.
    ^dib!
aborted
	"The scanning has been aborted.
	Currently do nothing."!   
fromHandle: hDIB
    "Answer an instance of Bitmap copied from a memory dib handle hDIB.
    Implemented by Ale..."

    | addressPtr address bitmapInfo bits dib rgb bytes |
    #addedByOStyle.
    (Smalltalk includesKey: #ExternalAddress)
	ifTrue: [	"32 bits version"
	    ( addressPtr :=  KernelLibrary globalLock: hDIB asParameter) = 0 ifTrue: [ ^self osError ].
    	address := (Smalltalk at: #ExternalAddress) fromInteger: addressPtr.
	]
	ifFalse: [	"16 bits version"
	    addressPtr := KernelLibrary globalLock: hDIB asParameter.
    	address := (Smalltalk at: #WinAddress) fromString: addressPtr.
    ].
    bitmapInfo := WinBitmapInfoHeader new.
    bitmapInfo contents
        replaceFrom: 1
        to: bitmapInfo contents size
        with: address
        startingAt: 1.

    bitmapInfo bitCount = 24
    ifFalse: [
        rgb := ByteArray new:  (
            WinBitmapInfo
                colorTableSize: bitmapInfo bitCount
                clrUsed: bitmapInfo clrUsed).
        rgb
            replaceFrom: 1
            to: rgb size
            with: address
            startingAt: bitmapInfo contents size + 1.
    ].

    bitmapInfo := WinBitmapInfo fromBytes: (
        rgb isNil
            ifTrue: [ bitmapInfo contents ]
            ifFalse: [ bitmapInfo contents, rgb ] ).
    bitmapInfo sizeImage = 0
    ifTrue: [
        bytes := (((bitmapInfo width * bitmapInfo bitCount) + 31) // 32 * 4) * bitmapInfo height
    ]
    ifFalse: [bytes := bitmapInfo sizeImage].

    bits := ByteArray new: bytes.
    bits
        replaceFrom: 1
        to: bits size
        with: address
        startingAt: bitmapInfo contents size + 1.

    KernelLibrary globalUnlock: hDIB asParameter.
    ^self new
        bitmapInfo: bitmapInfo;
        bits: bits;
        yourself!